@@ -51,7 +51,7 @@ bool With_clause::add_with_element(With_element *elem)
   elem->owner= this;
   elem->number= with_list.elements;
   elem->spec->with_element= elem;
-  with_list.link_in_list(elem, &elem->next);
+  with_list.insert(elem, &elem->next);
   return false;
 }
 
@@ -100,6 +100,19 @@ bool LEX::check_dependencies_in_with_clauses()
 }
 
 
+bool LEX::prepare_unreferenced_in_with_clauses()
+{
+  for (With_clause *with_clause= with_clauses_list;
+       with_clause;
+       with_clause= with_clause->next_with_clause)
+  {
+    if (with_clause->prepare_unreferenced_elements(thd))
+      return true;
+  }
+  return false;
+}
+
+
 /**
   @brief
     Resolve table references to CTE from a sub-chain of table references
@@ -153,6 +166,18 @@ bool LEX::resolve_references_to_cte(TABLE_LIST *tables,
 {
   With_element *with_elem= 0;
 
+  /*
+    Add back the tables collected during definition of the CTEs, but cleared
+    during mysql_init_delete
+  */
+  if (save_list.elements)
+  {
+    for (TABLE_LIST *saved= save_list.first; saved; saved= saved->next_global)
+      add_to_query_tables(saved);
+    save_list.empty();
+    last_table()->next_global= nullptr;
+  }
+
   for (TABLE_LIST *tbl= tables; tbl != *tables_last; tbl= tbl->next_global)
   {
     if (tbl->derived)
@@ -291,8 +316,7 @@ bool With_clause::check_dependencies()
          elem != with_elem;
          elem= elem->next)
     {
-      if (lex_string_cmp(system_charset_info, with_elem->get_name(),
-                         elem->get_name()) == 0)
+      if (with_elem->get_name().streq(elem->get_name()))
       {
         my_error(ER_DUP_QUERY_NAME, MYF(0),
                  with_elem->get_name_str());
@@ -359,7 +383,7 @@ struct st_unit_ctxt_elem
   @details
     For each table reference ref(T) from the FROM list of every select sl 
     immediately contained in the specification query of this element this
-    method searches for the definition of T in the the with clause which
+    method searches for the definition of T in the with clause which
     this element belongs to. If such definition is found then the dependency
     on it is set in sl->with_dep and in this->base_dep_map.  
 */
@@ -414,8 +438,7 @@ With_element *With_clause::find_table_def(TABLE_LIST *table,
   {
     if (excl_spec && with_elem->spec == excl_spec)
       continue;
-    if (my_strcasecmp(system_charset_info, with_elem->get_name_str(),
-                      table->table_name.str) == 0 &&
+    if (with_elem->get_name().streq(table->table_name) &&
         !table->is_fqtn)
     {
       table->set_derived();
@@ -527,7 +550,8 @@ void With_element::check_dependencies_in_select(st_select_lex *sl,
     if (is_spec_select)
     {
       With_clause *with_clause= sl->master_unit()->with_clause;
-      if (with_clause)
+      /* Non-recursive CTE cannot SELECT from itself */
+      if (with_clause && with_clause->with_recursive)
         tbl->with= with_clause->find_table_def(tbl, NULL, NULL);
       if (!tbl->with)
         tbl->with= owner->find_table_def(tbl,
@@ -624,7 +648,7 @@ TABLE_LIST *With_element::find_first_sq_rec_ref_in_select(st_select_lex *sel)
   @param  dep_map  IN/OUT The bit where to mark the found dependencies
 
   @details
-    This method searches in the unit 'unit' for the the references in FROM
+    This method searches in the unit 'unit' for the references in FROM
     lists of all selects contained in this unit and in the with clause
     attached to this unit that refer to definitions of tables from the
     same with clause as this element.
@@ -668,7 +692,7 @@ void With_element::check_dependencies_in_unit(st_select_lex_unit *unit,
   @param  dep_map      IN/OUT The bit where to mark the found dependencies
 
   @details
-    This method searches in the with_clause for the the references in FROM
+    This method searches in the with_clause for the references in FROM
     lists of all selects contained in the specifications of the with elements
     from this with_clause that refer to definitions of tables from the
     same with clause as this element.
@@ -694,7 +718,7 @@ With_element::check_dependencies_in_with_clause(With_clause *with_clause,
 
 /**
   @brief
-    Find mutually recursive with elements and check that they have ancors
+    Find mutually recursive with elements and check that they have anchors
  
   @details
     This method performs the following:
@@ -823,7 +847,7 @@ bool With_clause::check_anchors()
 	    el->work_dep_map|= elem->work_dep_map;          
         }
       }
-      /* If the transitive closure displays any cycle report an arror */
+      /* If the transitive closure displays any cycle report an error */
       elem= with_elem;
       while ((elem= elem->get_next_mutually_recursive()) != with_elem)
       {
@@ -1338,7 +1362,7 @@ With_element *st_select_lex::find_table_def_in_with_clauses(TABLE_LIST *table,
     /* 
       If sl->master_unit() is the spec of a with element then the search for 
       a definition was already done by With_element::check_dependencies_in_spec
-      and it was unsuccesful. Yet for units cloned from the spec it has not 
+      and it was unsuccessful. Yet for units cloned from the spec it has not
       been done yet.
     */
     With_clause *attached_with_clause= sl->get_with_clause();
@@ -1441,7 +1465,7 @@ bool st_select_lex::check_unrestricted_recursive(bool only_standard_compliant)
   if (!with_elem ||!with_elem->is_recursive)
   {
     /*
-      If this select is not from the specifiocation of a with elememt or
+      If this select is not from the specification of a with elememt or
       if this not a recursive with element then there is nothing to check.
     */
     return false;
@@ -1654,10 +1678,14 @@ void With_clause::print(THD *thd, String *str, enum_query_type query_type)
 }
 
 
-static void list_strlex_print(THD *thd, String *str, List<Lex_ident_sys> *list)
+void list_strlex_print(THD *thd, String *str, List<Lex_ident_sys> *list,
+                              bool bracketed)
 {
   List_iterator_fast<Lex_ident_sys> li(*list);
   bool first= TRUE;
+
+  if (bracketed)
+    str->append('(');
   while(Lex_ident_sys *col_name= li++)
   {
     if (first)
@@ -1666,6 +1694,8 @@ static void list_strlex_print(THD *thd, String *str, List<Lex_ident_sys> *list)
       str->append(',');
     append_identifier(thd, str, col_name);
   }
+  if (bracketed)
+    str->append(')');
 }
 
 
@@ -1686,12 +1716,7 @@ void With_element::print(THD *thd, String *str, enum_query_type query_type)
 {
   str->append(get_name());
   if (column_list.elements)
-  {
-    List_iterator_fast<Lex_ident_sys> li(column_list);
-    str->append('(');
-    list_strlex_print(thd, str, &column_list);
-    str->append(')');
-  }
+    list_strlex_print(thd, str, &column_list, true);
   str->append(STRING_WITH_LEN(" as ("));
   spec->print(str, query_type);
   str->append(')');
