@@ -52,8 +52,7 @@
 #define MAX_DROP_TABLE_Q_LEN      1024
 
 const char *del_exts[]= {".BAK", ".opt", NullS};
-static TYPELIB deletable_extensions=
-{array_elements(del_exts)-1,"del_exts", del_exts, NULL};
+static TYPELIB deletable_extensions= CREATE_TYPELIB_FOR(del_exts);
 
 static bool find_db_tables_and_rm_known_files(THD *, MY_DIR *,
                                               const Lex_ident_db_normalized &db,
@@ -61,10 +60,8 @@ static bool find_db_tables_and_rm_known_files(THD *, MY_DIR *,
 
 long mysql_rm_arc_files(THD *thd, MY_DIR *dirp, const char *org_path);
 my_bool rm_dir_w_symlink(const char *org_path, my_bool send_error);
-static void mysql_change_db_impl(THD *thd,
-                                 LEX_CSTRING *new_db_name,
-                                 privilege_t new_db_access,
-                                 CHARSET_INFO *new_db_charset);
+static void mysql_change_db_impl(THD *, LEX_CSTRING *, privilege_t,
+                                 CHARSET_INFO *);
 static bool mysql_rm_db_internal(THD *thd, const Lex_ident_db &db,
                                  bool if_exists, bool silent);
 
@@ -98,12 +95,10 @@ typedef struct my_dbopt_st
 */
 
 static inline bool
-cmp_db_names(LEX_CSTRING *db1_name, const LEX_CSTRING *db2_name)
+cmp_db_names(const Lex_ident_db &db1_name, const Lex_ident_db &db2_name)
 {
-  return (db1_name->length == db2_name->length &&
-          (db1_name->length == 0 ||
-           my_strcasecmp(table_alias_charset,
-                         db1_name->str, db2_name->str) == 0));
+  return (db1_name.length == 0 && db2_name.length == 0) ||
+         db1_name.streq(db2_name);
 }
 
 #ifdef HAVE_PSI_INTERFACE
@@ -147,16 +142,17 @@ private:
   Hash_set<LEX_STRING> m_set;
   mysql_rwlock_t m_lock;
 
-  static uchar *get_key(const LEX_STRING *ls, size_t *sz, my_bool)
+  static const uchar *get_key(const void *ls_, size_t *sz, my_bool)
   {
+    const LEX_STRING *ls= static_cast<const LEX_STRING*>(ls_);
     *sz= ls->length;
-    return (uchar *) ls->str;
+    return reinterpret_cast<const uchar*>(ls->str);
   }
 
 public:
   dbname_cache_t()
       : m_set(key_memory_dbnames_cache, table_alias_charset, 10, 0,
-              sizeof(char *), (my_hash_get_key) get_key, my_free, 0)
+              sizeof(char *), get_key, my_free, 0)
   {
     mysql_rwlock_init(key_rwlock_LOCK_dbnames, &m_lock);
   }
@@ -240,14 +236,14 @@ static int my_rmdir(const char *dir)
   Function we use in the creation of our hash to get key.
 */
 
-extern "C" uchar* dboptions_get_key(my_dbopt_t *opt, size_t *length,
-                                    my_bool not_used);
+extern "C" const uchar *dboptions_get_key(const void *opt, size_t *length,
+                                          my_bool);
 
-uchar* dboptions_get_key(my_dbopt_t *opt, size_t *length,
-                         my_bool not_used __attribute__((unused)))
+const uchar *dboptions_get_key(const void *opt_, size_t *length, my_bool)
 {
+  auto opt= static_cast<const my_dbopt_t *>(opt_);
   *length= opt->name_length;
-  return (uchar*) opt->name;
+  return reinterpret_cast<const uchar *>(opt->name);
 }
 
 
@@ -299,8 +295,8 @@ bool my_dboptions_cache_init(void)
   {
     dboptions_init= 1;
     error= my_hash_init(key_memory_dboptions_hash, &dboptions,
-                        table_alias_charset, 32, 0, 0, (my_hash_get_key)
-                        dboptions_get_key, free_dbopt, 0);
+                        table_alias_charset, 32, 0, 0, dboptions_get_key,
+                        free_dbopt, 0);
   }
   dbname_cache_init();
   return error;
@@ -333,7 +329,7 @@ void my_dbopt_cleanup(void)
   mysql_rwlock_wrlock(&LOCK_dboptions);
   my_hash_free(&dboptions);
   my_hash_init(key_memory_dboptions_hash, &dboptions, table_alias_charset, 32,
-               0, 0, (my_hash_get_key) dboptions_get_key, free_dbopt, 0);
+               0, 0, dboptions_get_key, free_dbopt, 0);
   mysql_rwlock_unlock(&LOCK_dboptions);
 }
 
@@ -536,36 +532,53 @@ static bool write_db_opt(THD *thd, const char *path,
 
   DESCRIPTION
 
+  create->default_table_charset is guaranteed to be alway set
+  Required by some callers
+
   RETURN VALUES
   0	File found
-  1	No database file or could not open it
-
+  -1	No database file (file was not found or 'empty' file was cached)
+  1     Could not open it
 */
 
-bool load_db_opt(THD *thd, const char *path, Schema_specification_st *create)
+int load_db_opt(THD *thd, const char *path, Schema_specification_st *create)
 {
   File file;
   char buf[256+DATABASE_COMMENT_MAXLEN];
   DBUG_ENTER("load_db_opt");
-  bool error=1;
+  int error= 0;
   size_t nbytes;
   myf utf8_flag= thd->get_utf8_flag();
 
   bzero((char*) create,sizeof(*create));
-  create->default_table_charset= thd->variables.collation_server;
 
   /* Check if options for this database are already in the hash */
   if (!get_dbopt(thd, path, create))
-    DBUG_RETURN(0);
+  {
+    if (!create->default_table_charset)
+      error= -1;                                // db.opt did not exists
+    goto err1;
+  }
 
   /* Otherwise, load options from the .opt file */
   if ((file= mysql_file_open(key_file_dbopt,
                              path, O_RDONLY | O_SHARE, MYF(0))) < 0)
+  {
+    /*
+      Create an empty entry, to avoid doing an extra file open for every create
+      table.
+    */
+    put_dbopt(path, create);
+    error= -1;
     goto err1;
+  }
 
   IO_CACHE cache;
   if (init_io_cache(&cache, file, IO_SIZE, READ_CACHE, 0, 0, MYF(0)))
-    goto err2;
+  {
+    error= 1;
+    goto err2;                                  // Not cached
+  }
 
   while ((int) (nbytes= my_b_gets(&cache, (char*) buf, sizeof(buf))) > 0)
   {
@@ -586,12 +599,12 @@ bool load_db_opt(THD *thd, const char *path, Schema_specification_st *create)
            default-collation commands.
         */
         if (!(create->default_table_charset=
-        get_charset_by_csname(pos+1, MY_CS_PRIMARY, MYF(utf8_flag))) &&
+              get_charset_by_csname(pos+1, MY_CS_PRIMARY, MYF(utf8_flag))) &&
             !(create->default_table_charset=
               get_charset_by_name(pos+1, MYF(utf8_flag))))
         {
           sql_print_error("Error while loading database options: '%s':",path);
-          sql_print_error(ER_THD(thd, ER_UNKNOWN_CHARACTER_SET),pos+1);
+          sql_print_error(ER_DEFAULT(ER_UNKNOWN_CHARACTER_SET),pos+1);
           create->default_table_charset= default_charset_info;
         }
       }
@@ -600,7 +613,7 @@ bool load_db_opt(THD *thd, const char *path, Schema_specification_st *create)
         if (!(create->default_table_charset= get_charset_by_name(pos+1, MYF(utf8_flag))))
         {
           sql_print_error("Error while loading database options: '%s':",path);
-          sql_print_error(ER_THD(thd, ER_UNKNOWN_COLLATION),pos+1);
+          sql_print_error(ER_DEFAULT(ER_UNKNOWN_COLLATION),pos+1);
           create->default_table_charset= default_charset_info;
         }
       }
@@ -621,10 +634,11 @@ bool load_db_opt(THD *thd, const char *path, Schema_specification_st *create)
 err2:
   mysql_file_close(file, MYF(0));
 err1:
+  if (!create->default_table_charset)           // In case of error
+    create->default_table_charset= thd->variables.collation_server;
   DBUG_RETURN(error);
 }
 
-
 /*
   Retrieve database options by name. Load database options file or fetch from
   cache.
@@ -651,11 +665,12 @@ err1:
     db_create_info right after that.
 
   RETURN VALUES (read NOTE!)
-    FALSE   Success
-    TRUE    Failed to retrieve options
+  0	File found
+  -1	No database file (file was not found or 'empty' file was cached)
+  1     Could not open it
 */
 
-bool load_db_opt_by_name(THD *thd, const char *db_name,
+int load_db_opt_by_name(THD *thd, const char *db_name,
                          Schema_specification_st *db_create_info)
 {
   char db_opt_path[FN_REFLEN + 1];
@@ -810,7 +825,7 @@ mysql_create_db_internal(THD *thd, const Lex_ident_db &db,
     /*
       We come here when we managed to create the database, but not the option
       file.  In this case it's best to just continue as if nothing has
-      happened.  (This is a very unlikely senario)
+      happened.  (This is a very unlikely scenario)
     */
     thd->clear_error();
   }
@@ -1225,7 +1240,7 @@ update_binlog:
     char *query, *query_pos, *query_end, *query_data_start;
     TABLE_LIST *tbl;
 
-    if (!(query= (char*) thd->alloc(MAX_DROP_TABLE_Q_LEN)))
+    if (!(query= thd->alloc(MAX_DROP_TABLE_Q_LEN)))
       goto exit; /* not much else we can do */
     query_pos= query_data_start= strmov(query,"DROP TABLE IF EXISTS ");
     query_end= query + MAX_DROP_TABLE_Q_LEN;
@@ -1239,7 +1254,7 @@ update_binlog:
       if (ha_table_exists(thd, &tbl->db, &tbl->table_name))
         continue;
 
-      tbl_name_len= my_snprintf(quoted_name, sizeof(quoted_name), "%`s",
+      tbl_name_len= my_snprintf(quoted_name, sizeof(quoted_name), "%sQ",
                                 tbl->table_name.str);
       tbl_name_len++;                           /* +1 for the comma */
       if (query_pos + tbl_name_len + 1 >= query_end)
@@ -1290,7 +1305,8 @@ exit:
     SELECT DATABASE() in the future). For this we free() thd->db and set
     it to 0.
   */
-  if (unlikely(thd->db.str && cmp_db_names(&thd->db, &db) && !error))
+  if (unlikely(thd->db.str &&
+               cmp_db_names(Lex_ident_db(thd->db), db) && !error))
   {
     mysql_change_db_impl(thd, NULL, NO_ACL, thd->variables.collation_server);
     thd->session_tracker.current_schema.mark_as_changed(thd);
@@ -1331,25 +1347,24 @@ static bool find_db_tables_and_rm_known_files(THD *thd, MY_DIR *dirp,
 
   for (size_t idx=0; idx < files.elements(); idx++)
   {
-    LEX_CSTRING *table= files.at(idx);
+    const LEX_CSTRING *table= files.at(idx);
 
     /* Drop the table nicely */
-    TABLE_LIST *table_list=(TABLE_LIST*)thd->calloc(sizeof(*table_list));
+    TABLE_LIST *table_list= thd->calloc<TABLE_LIST>(1);
 
     if (!table_list)
       DBUG_RETURN(true);
     table_list->db= db;
-    table_list->table_name= *table;
-    table_list->open_type= OT_BASE_ONLY;
-
     /*
       On the case-insensitive file systems table is opened
       with the lowercased file name. So we should lowercase
       as well to look up the cache properly.
     */
-    if (lower_case_file_system)
-      table_list->table_name.length= my_casedn_str(files_charset_info,
-                                                   (char*) table_list->table_name.str);
+    table_list->table_name= lower_case_file_system ?
+                            Lex_ident_table(thd->make_ident_casedn(*table)) :
+                            Lex_ident_table(*table);
+
+    table_list->open_type= OT_BASE_ONLY;
 
     table_list->alias= table_list->table_name;	// If lower_case_table_names=2
     MDL_REQUEST_INIT(&table_list->mdl_request, MDL_key::TABLE,
@@ -1416,7 +1431,7 @@ static bool find_db_tables_and_rm_known_files(THD *thd, MY_DIR *dirp,
 
   SYNOPSIS
     rm_dir_w_symlink()
-    org_path    path of derictory
+    org_path    path of directory
     send_error  send errors
   RETURN
     0 OK
@@ -1548,8 +1563,7 @@ err:
   @param new_db_charset Character set of the new database.
 */
 
-static void mysql_change_db_impl(THD *thd,
-                                 LEX_CSTRING *new_db_name,
+static void mysql_change_db_impl(THD *thd, LEX_CSTRING *new_db_name,
                                  privilege_t new_db_access,
                                  CHARSET_INFO *new_db_charset)
 {
@@ -1613,8 +1627,7 @@ static void mysql_change_db_impl(THD *thd,
                                   "length" is set to 0.
 */
 
-static void backup_current_db_name(THD *thd,
-                                   LEX_STRING *saved_db_name)
+static void backup_current_db_name(THD *thd, LEX_STRING *saved_db_name)
 {
   DBUG_ASSERT(saved_db_name->length >= SAFE_NAME_LEN +1);
   if (!thd->db.str)
@@ -1636,18 +1649,18 @@ static void backup_current_db_name(THD *thd,
 
   @param thd          thread handle
   @param new_db_name  database name
-  @param force_switch if force_switch is FALSE, then the operation will fail if
+  @param force        if force is FALSE, then the operation will fail if
 
                         - new_db_name is NULL or empty;
 
                         - OR new database name is invalid
-                          (check_db_name() failed);
+                          (Lex_ident_db::check_name() failed);
 
                         - OR user has no privilege on the new database;
 
                         - OR new database does not exist;
 
-                      if force_switch is TRUE, then
+                      if force is TRUE, then
 
                         - if new_db_name is NULL or empty, the current
                           database will be NULL, @@collation_database will
@@ -1655,8 +1668,9 @@ static void backup_current_db_name(THD *thd,
                           succeed.
 
                         - if new database name is invalid
-                          (check_db_name() failed), the current database
-                          will be NULL, @@collation_database will be set to
+                          (Lex_ident_db::check_name() failed),
+                          the current database will be NULL,
+                          @@collation_database will be set to
                           @@collation_server, but the operation will fail;
 
                         - user privileges will not be checked
@@ -1693,8 +1707,7 @@ static void backup_current_db_name(THD *thd,
     @retval >0  Error
 */
 
-uint mysql_change_db(THD *thd, const LEX_CSTRING *new_db_name,
-                     bool force_switch)
+uint mysql_change_db(THD *thd, const LEX_CSTRING &new_db_name, bool force)
 {
   DBNameBuffer new_db_buff;
   LEX_CSTRING new_db_file_name;
@@ -1704,9 +1717,9 @@ uint mysql_change_db(THD *thd, const LEX_CSTRING *new_db_name,
   CHARSET_INFO *db_default_cl;
   DBUG_ENTER("mysql_change_db");
 
-  if (new_db_name->length == 0)
+  if (new_db_name.length == 0)
   {
-    if (force_switch)
+    if (force)
     {
       /*
         This can happen only if we're switching the current database back
@@ -1728,9 +1741,9 @@ uint mysql_change_db(THD *thd, const LEX_CSTRING *new_db_name,
       DBUG_RETURN(ER_NO_DB_ERROR);
     }
   }
-  DBUG_PRINT("enter",("name: '%s'", new_db_name->str));
+  DBUG_PRINT("enter",("name: '%s'", new_db_name.str));
 
-  if (is_infoschema_db(new_db_name))
+  if (is_infoschema_db(&new_db_name))
   {
     /* Switch the current database to INFORMATION_SCHEMA. */
 
@@ -1741,11 +1754,12 @@ uint mysql_change_db(THD *thd, const LEX_CSTRING *new_db_name,
 
   new_db_file_name= lower_case_table_names == 1 ?
                     new_db_buff.copy_casedn(&my_charset_utf8mb3_general_ci,
-                                            *new_db_name).to_lex_cstring() :
-                    *new_db_name;
+                                            new_db_name).to_lex_cstring() :
+                    new_db_name;
 
   /*
-    NOTE: if check_db_name() fails, we should throw an error in any case,
+    NOTE: if Lex_ident_db::check_name() fails,
+    we should throw an error in any case,
     even if we are called from sp_head::execute().
 
     It's next to impossible however to get this error when we are called
@@ -1754,9 +1768,9 @@ uint mysql_change_db(THD *thd, const LEX_CSTRING *new_db_name,
     The cast below ok here as new_db_file_name was just allocated
   */
 
-  if (Lex_ident_fs(new_db_file_name).check_db_name_with_error())
+  if (Lex_ident_db::check_name_with_error(new_db_file_name))
   {
-    if (force_switch)
+    if (force)
       mysql_change_db_impl(thd, NULL, NO_ACL, thd->variables.collation_server);
 
     DBUG_RETURN(ER_WRONG_DB_NAME);
@@ -1775,8 +1789,7 @@ uint mysql_change_db(THD *thd, const LEX_CSTRING *new_db_name,
     db_access|= sctx->master_access;
   }
 
-  if (!force_switch &&
-      !(db_access & DB_ACLS) &&
+  if (!force&& !(db_access & DB_ACLS) &&
       check_grant_db(thd, new_db_file_name.str))
   {
     my_error(ER_DBACCESS_DENIED_ERROR, MYF(0),
@@ -1793,7 +1806,7 @@ uint mysql_change_db(THD *thd, const LEX_CSTRING *new_db_name,
 
   if (check_db_dir_existence(new_db_file_name.str))
   {
-    if (force_switch)
+    if (force)
     {
       /* Throw a warning and free new_db_file_name. */
 
@@ -1828,8 +1841,7 @@ uint mysql_change_db(THD *thd, const LEX_CSTRING *new_db_name,
     NOTE: in mysql_change_db_impl() new_db_file_name is assigned to THD
     attributes and will be freed in THD::~THD().
   */
-  if (const char *tmp= my_strndup(key_memory_THD_db,
-                                  new_db_file_name.str,
+  if (const char *tmp= my_strndup(key_memory_THD_db, new_db_file_name.str,
                                   new_db_file_name.length, MYF(MY_WME)))
   {
     LEX_CSTRING new_db_malloced({tmp, new_db_file_name.length});
@@ -1849,8 +1861,8 @@ done:
   Change the current database and its attributes if needed.
 
   @param          thd             thread handle
-  @param          new_db_name     database name
-  @param[in, out] saved_db_name   IN: "str" points to a buffer where to store
+  @param          new_db          database name
+  @param[in, out] saved_db        IN: "str" points to a buffer where to store
                                   the old database name, "length" contains the
                                   buffer size
                                   OUT: if the current (default) database is
@@ -1865,20 +1877,18 @@ done:
                                   the function suceeded)
 */
 
-bool mysql_opt_change_db(THD *thd,
-                         const LEX_CSTRING *new_db_name,
-                         LEX_STRING *saved_db_name,
-                         bool force_switch,
+bool mysql_opt_change_db(THD *thd, const LEX_CSTRING &new_db,
+                         LEX_STRING *saved_db, bool force_switch,
                          bool *cur_db_changed)
 {
-  *cur_db_changed= !cmp_db_names(&thd->db, new_db_name);
+  *cur_db_changed= !cmp_db_names(Lex_ident_db(thd->db), Lex_ident_db(new_db));
 
   if (!*cur_db_changed)
     return FALSE;
 
-  backup_current_db_name(thd, saved_db_name);
+  backup_current_db_name(thd, saved_db);
 
-  return mysql_change_db(thd, new_db_name, force_switch);
+  return mysql_change_db(thd, new_db, force_switch);
 }
 
 
@@ -1953,9 +1963,8 @@ bool mysql_upgrade_db(THD *thd, const Lex_ident_db &old_db)
   }
 
   /* Step1: Create the new database */
-  if (unlikely((error= mysql_create_db_internal(thd, new_db,
-                                                DDL_options(), &create_info,
-                                                1))))
+  if (unlikely((error= mysql_create_db_internal(thd, new_db, DDL_options(),
+                                                &create_info, 1))))
     goto exit;
 
   /* Step2: Move tables to the new database */
@@ -1969,7 +1978,7 @@ bool mysql_upgrade_db(THD *thd, const Lex_ident_db &old_db)
       LEX_CSTRING table_str;
       DBUG_PRINT("info",("Examining: %s", file->name));
 
-      /* skiping non-FRM files */
+      /* skipping non-FRM files */
       if (!(extension= (char*) fn_frm_ext(file->name)))
         continue;
 
@@ -2009,10 +2018,9 @@ bool mysql_upgrade_db(THD *thd, const Lex_ident_db &old_db)
       old database and some tables in the new database.
       Let's delete the option file, and then the new database directory.
       If some tables were left in the new directory, rmdir() will fail.
-      It garantees we never loose any tables.
+      It guarantees we never lose any tables.
     */
-    build_table_filename(path, sizeof(path)-1,
-                         new_db.str,"",MY_DB_OPT_FILE, 0);
+    build_table_filename(path, sizeof(path)-1, new_db.str,"",MY_DB_OPT_FILE, 0);
     mysql_file_delete(key_file_dbopt, path, MYF(MY_WME));
     length= build_table_filename(path, sizeof(path)-1, new_db.str, "", "", 0);
     if (length && path[length-1] == FN_LIBCHAR)
@@ -2058,8 +2066,9 @@ bool mysql_upgrade_db(THD *thd, const Lex_ident_db &old_db)
       char oldname[FN_REFLEN + 1], newname[FN_REFLEN + 1];
       DBUG_PRINT("info",("Examining: %s", file->name));
 
-      /* skiping MY_DB_OPT_FILE */
-      if (!my_strcasecmp(files_charset_info, file->name, MY_DB_OPT_FILE))
+      /* skipping MY_DB_OPT_FILE */
+      if (!files_charset_info->strnncoll(Lex_cstring_strlen(file->name),
+                                         Lex_cstring_strlen(MY_DB_OPT_FILE)))
         continue;
 
       /* pass empty file name, and file->name as extension to avoid encoding */
@@ -2092,7 +2101,7 @@ bool mysql_upgrade_db(THD *thd, const Lex_ident_db &old_db)
 
   /* Step9: Let's do "use newdb" if we renamed the current database */
   if (change_to_newdb)
-    error|= mysql_change_db(thd, & new_db, FALSE) != 0;
+    error|= mysql_change_db(thd, new_db, FALSE) != 0;
 
 exit:
   DBUG_RETURN(error);
